function    [hat_Q_vec, diff_vec] = hat_Q_calculation( S  )
%   USAGE:   simulation_results = simulation_individual(  S, T_sim, run_shock,  seed   )
%   S includes the post processing results. 
%   simulation_results is a struct of variables after simulation.

%% SECTION 1. Simulation
dt = S.dt; w_grid = S.w_grid; 
hat_Q_vec = zeros( size(w_grid) );  
[~,epsi_nodes,weight_nodes] = GH_Quadrature(10,1,1);
spread = S.rf_vec - S.rg_vec;
r_vec = S.rh_vec;

tol = 10^(-5); max_it=1000;  
diff=1;  diff_vec = [];  counter=0;
while( diff>tol && counter<max_it )
    hat_Q_vec_last = hat_Q_vec;
    hat_Q_fitted = interp1( w_grid, hat_Q_vec, 'linear', 'pp' );
    for( i = 2: numel(w_grid))
        w = w_grid(i);
        muw =  ppval(S.muw_fitted, w) ;
        sigmaw = ppval(S.sigmaw_fitted, w );
        kappaw = ppval(S.kappaw_fitted , w ); 
        w_noshock = w*(1 + muw*dt + sigmaw*epsi_nodes );
        w_shock = w*(1 + muw*dt + sigmaw*epsi_nodes - kappaw );
        expectation = ( 1-S.lambda*dt )  * weight_nodes' * ppval( hat_Q_fitted,  w_noshock ) + S.lambda*dt * weight_nodes' * ppval( hat_Q_fitted,  w_shock );
        hat_Q_vec(i) =   S.Q(w) * spread(i)*dt  + exp(- r_vec(i)*dt) * expectation ; 
    end
    hat_Q_vec(1) = S.Q(0)* spread(1)/r_vec(1);
    diff = sum(abs(hat_Q_vec-hat_Q_vec_last));
    diff_vec = [diff_vec, diff];
    counter = counter + 1;
end

