function   [ xK, xg, kappap, exit_flag_xK, exit_flag_xg, exit_flag_kappap ] = inside_iteration( w, p, p_fitted, xK, xg, pw, i, print_level, par )
% Method: For each kappap, solve xK. Then get a function of xK(kappap) and solve kappap. Finally update xg.
% This function returns a newly solved xK, xg and kappap within an functional iteration loop.

AH = par.AH;   AL=par.AL;  sigmaK=par.sigmaK; investment=par.investment;  investmentD=par.investmentD;  Q=par.Q;  hat_Q=par.hat_Q; 
beta=par.beta;  alpha0=par.alpha0;  lambda=par.lambda;   e=par.e;   theta=par.theta;  max_jump=par.max_jump;   rho=par.rho;
v_prime_fun = par.v_prime_fun;   p0=par.p0;   p1=par.p1;   

% ======== Step 1: Start with last round p and xg, and define commonly used functions. ========
yK_fun = @(xK)  (p/(p+hat_Q) -w*xK)./(1-w);
yg_fun = @(xg)  (Q(w)/(p+hat_Q) - w*xg) / (1-w)  ;  yg=yg_fun(xg);
sigmap_fun = @(xK)  pw*w*(1-w)*(xK-yK_fun(xK)) ./ ( 1- pw*w*(1-w)*(xK-yK_fun(xK))  ) * sigmaK;
delta_x_fun = @(xK,xg)  max(beta*(xK+xg-1) - xg, 0) ;
indicator_fun = @(xK,xg)  beta*(xK+xg-1) - xg>0;
kappad_fun = @(xK,xg)  theta * max(xK+e-1,0)./(xK+xg-1);
kappa_fs_fun = @(xK, xg) alpha0/(1-alpha0)*delta_x_fun(xK,xg)*w/(1-w);

%======== Step 2:  Solve the xK and kappap together. ========
xK_upper = min( [  (1 + 1./(w*pw))*(pw>=0) + 10^10*(pw<0) ,  p./(p+hat_Q)./w  ] );    % Note: pw might go negative in some scenario. Although we need to eliminate this case.
xK_lower =  1; 
kappap_upper = max_jump;
kappap_lower = 0;

kappab_fun = @(kappap, xK)  theta*(1-e) + (1-theta)*( xK*kappap + alpha0/(1-alpha0)*delta_x_fun(xK,xg) ) ; 
kappah_fun = @(kappap, xK) yK_fun(xK).*kappap + (1-yK_fun(xK)-yg)*kappad_fun(xK,xg)  - kappa_fs_fun(xK,xg) ;

kappa_residual = @(kappap, xK)  p - ppval( w.*(1-kappab_fun(kappap,xK))./(1-kappah_fun(kappap,xK)-w.*(kappab_fun(kappap,xK)-kappah_fun(kappap,xK))) , p_fitted ) - kappap ...
    + 10^30* ( max(kappab_fun(kappap,xK)-1,0) + max(kappah_fun(kappap,xK)-1,0) + max(kappap-kappap_upper,0) + abs(min(kappap-kappap_lower,0)) )  ;    
xK_residual = @(kappap,xK) 10^3 * ( ( sigmaK+sigmap_fun(xK) ).^2 .* ( xK - yK_fun(xK) ) ... 
                              +  lambda*(1-theta)*( (kappap+ indicator_fun(xK,xg).*alpha0/(1-alpha0)*beta)./( 1-xK.*kappap-alpha0./(1-alpha0).*delta_x_fun(xK,xg) ) ) ...
                              -  lambda*( kappap - kappad_fun(xK,xg) ) ./ (  1-yK_fun(xK).*kappap-(1-yK_fun(xK)-yg).*kappad_fun(xK,xg) + kappa_fs_fun(xK,xg) )    - (AH-AL)/p )   ...
                              + 10^20*(  max( xK-xK_upper,0 ) + abs( min( xK-xK_lower, 0) ) +  max(alpha0./(1-alpha0).*delta_x_fun(xK,xg)+xK*kappap-1,0) );   % Penalty for xceeding the limits                            
exit_flag_xK=0; exit_flag_kappap=0;    exit_flag_xg=0;

% Test
if( print_level>=2 )
    upper = (  1-alpha0/(1-alpha0)*( beta*(xg-1)-xg ) ) ./ ( 0 + alpha0/(1-alpha0*beta) );
    xK_test_vec = linspace(  xK_lower, upper,  100 )' ;
    plot(  xK_test_vec, xK_residual( 0, xK_test_vec ) );   xlabel('x^K');   ylabel(['xK residual' ]);
    kp_test_vec = linspace(  kappap_lower, 0.012,  200 )' ;
    plot(  kp_test_vec, kappa_residual( kp_test_vec, xK )  );   xlabel('\kappa^p');   ylabel(['\kappa^p residual' ]);
end
% End of test

if(xK_upper>1)  % ************** Important:  Only consider the normal cases when xK>1 for the first part. 
    % Solve xK for different kappaps
    kappaps = [ 0; exp(linspace( log(0.001), log(max_jump), 19 )') ];   xKs = zeros(size(kappaps));   solved_xKs = zeros(size(kappaps)); 
    for( j = 1:numel(kappaps) )
        kappap = kappaps(j);
        xK_tigher_upper = ( 1-theta*(1-e)-(1-theta)*(alpha0/(1-alpha0))*( beta*(xg-1)-xg ) ) ./ ( (1-theta)*(kappap+alpha0/(1-alpha0)*beta) );
        [ xKs(j) , residual, ~ ] = fsolve( @(xK) xK_residual(kappap,xK) + 10^20*max(xK-xK_tigher_upper,0)  , 0.9*min([xK, (1-e)/kappap, xK_tigher_upper])  ,  optimset('Display','off' ) ); 
        if(  abs(residual) > 10^(-6) )
            if( abs(xKs(j)-p/(p+hat_Q)/w)<10^(-5) && xK_residual(kappap,xKs(j))<0  )   % If the residual is not small enough, check whether the solution hits the order of yK=0. 
                residual=0;
            end
            if( abs( xKs(j) - (1-beta)/beta*xg-1 ) < 10^(-6) )
                xKs(j) = (1-beta)/beta*xg+1;   % This is the just full liquidity insurance scenario. 
                residual = 0;
            end
        end
    %     % TEST
    %     xKs = linspace(  xK_lower,0.9* xK_tigher_upper, 100  )';
    %     plot( xKs,   xK_residual(kappap, xKs)  );
    %     % END OF TEST

        solved_xKs(j) = abs(residual)<10^(-6);
    end

    %         plot(  kappaps(solved_xKs==1), xKs(solved_xKs==1) )
    %         if( sum(solved_xKs~=1) >2 )
    %             disp( 'Inaccurate solution of xK(kappap)!' )
    %             plot( kappaps(solved_xKs==1), xKs(solved_xKs==1) );  pause(1);
    %         end

    % Solve kappap
    % roots_xK = interp1( kappaps,xKs,'linear','pp');
    % [ kappap , ~, exit_flag_kappap_xK ] = fsolve( @(kappap) kappa_residual(kappap, ppval(roots_xK, kappap) ) , 0.001 ,  optimset('Display','off' ) );
    % xK = ppval(roots_xK, kappap) ;

    roots_xK = griddedInterpolant( kappaps(solved_xKs==1),xKs(solved_xKs==1)  );
    if( sum( abs(xKs+xg-1)<10^(-6) )==0 )  % ******************** Banks still borrow from households.
        [ kappap , ~, exit_flag_kappap ] = fsolve( @(kappap) kappa_residual(kappap, roots_xK(kappap) ) , 0.001 ,  optimset('Display','off' ) );

        xK_tigher_upper = ( 1-theta*(1-e)-(1-theta)*(alpha0/(1-alpha0))*( beta*(xg-1)-xg ) ) ./ ( (1-theta)*(kappap+alpha0/(1-alpha0)*beta) );
        [ xK , ~, exit_flag_xK ] = fsolve( @(xK) xK_residual(kappap,xK) + 10^20*max(xK-xK_tigher_upper,0)  ,   roots_xK(kappap)    ,  optimset('Display','off' ) ); 
        if( abs( xK - p/(p+hat_Q)/w ) < 10^(-6) )   % At the corner of xK=1/w
            xK=p/(p+hat_Q)/w;  exit_flag_xK = xK_residual(kappap,xK)<0;
        end
        if( abs( xK - (1-beta)/beta*xg-1 ) < 10^(-6) )
            xK = (1-beta)/beta*xg+1;   % This is the just full liquidity insurance scenario. 
            exit_flag_xK = 1; 
        end

        % ======== Step 3: Solve xg. ========
        xg_upper = Q(w)/(w*(p+hat_Q));   
        xg_lower =  max( [ e,   1/(1-beta)*( beta*(xK-1) - (1-alpha0)/alpha0*(1-xK*kappap) )]  ) ;  
        % 1/w*max( Q(w)/(p+hat_Q) - (1-w), 0 ),   
        xg_residual = @(xg)  10^3 *( indicator_fun(xK,xg) * lambda.*(1-theta) * ( alpha0/(1-alpha0)*(1-beta) ) ./ ( 1-xK*kappap - alpha0/(1-alpha0)*delta_x_fun(xK,xg) )    ... 
                                       -  lambda*kappad_fun(xK,xg)./( 1 - yK_fun(xK).*kappap - (1-yK_fun(xK)-yg_fun(xg)).*kappad_fun(xK,xg) + kappa_fs_fun(xK,xg) )  - v_prime_fun( max(yg_fun(xg),0) )  ) ...
                                       + 10^20 * (max( xg-xg_upper,0 ) + abs( min( xg-xg_lower, 0) ))  ; % Penalty term
        % TEST
        if( print_level>=2 )
            xg_test_vec = linspace(  xg_lower, 0.97*xg_upper,  200 )' ;
            plot(  xg_test_vec, xg_residual( xg_test_vec )  );   xlabel('x^g');   ylabel(['xg residual' ]);
        end
        % END OF TEST
        [xg, ~, exit_flag_xg] = fsolve( xg_residual,  min(xg, 0.999*xg_upper) ,  optimset('Display','off' ));
        if(  abs( xg-beta/(1-beta)*(xK-1) )<10^(-6) )
            xg = beta/(1-beta)*(xK-1);   % Full liquidity insurance. 
            exit_flag_xg  = 1; 
        end

        if( abs(xg-e)<10^(-6)  &&  xg_residual(xg)<0)  % 1/w*max( Q(w)/(p+hat_Q) - (1-w), 0 )
            xg = e;   % xg reaching the boundary 
            exit_flag_xg =  xg_residual(xg)<0 ;
        end

        if( abs( xg - Q(w)/(w*(p+hat_Q)) ) < 10^(-3)  )
            xg = Q(w)/(w*(p+hat_Q));
            exit_flag_xg = 1;
        end

    end  % ******************** End of banks still borrow from hosueholds.
end   % **************** End of considering normal cases. 

% % ==== Step 5: Consider Special Cases ====
% if( exit_flag_kappap~=1 ||  exit_flag_xK~=1  || exit_flag_xg~=1 )  % Think about the special case of xK <= 1, which means no bank run at all. 
% %    disp( ['At i=', num2str(i), '(w=', num2str(w) ,') speical case of xK<=1 is considered' ] )
%    kappap = 0; 
%    kappab = theta*(1-e);
%    kappah = 0;  xg = 0;
%    kappap_test = ppval( p_fitted, w ) - ppval( p_fitted, w*(1-kappab)/(1-kappah-w*(kappab-kappah)) );
%    if( abs(kappap_test)<10^(-6) )   % If indeed the current w justifies a low risk taking and no bank run
%        xK_upper = min(p./(p+hat_Q)./w,1) ;  xK_lower=0;
%        xK_residual_norun = @(xK)  (AH-AL)/p - (sigmaK+sigmap_fun(xK)).^2.*(xK-yK_fun(xK)) - (xK+e<=1)*lambda*theta/(1-xK) + 10^20*( max( xK-xK_upper,0 ) + abs( min( xK-xK_lower, 0) )  ) ;
%        [ xK , residual , ~ ] = fsolve( xK_residual_norun,   1-e/2 ,  optimset('Display','off' ) );
%        if( abs(residual)<10^(-6) )
%            exit_flag_xK=1;  exit_flag_xg = 1;  exit_flag_kappap=1; 
%        elseif( abs(xK-xK_upper)<10^(-6) && xK_residual_norun(xK)>=0 )
%            xK = xK_upper; 
%            exit_flag_xK=1;  exit_flag_xg = 1;  exit_flag_kappap=1; 
%        end
%    end   
% end

% ==== Step 6: Collect values ====
% exit_flag_xK = exit_flag_kappap_xK;    exit_flag_kappap=exit_flag_kappap_xK;

if(exit_flag_kappap~=1 ||  exit_flag_xK~=1  || exit_flag_xg~=1)
    message = ['At i=', num2str(i), ', model not solved! Include' ];
    if( exit_flag_kappap~=1 )
        message = [ message, ' kappap' ];
    end
    if( exit_flag_xK~=1 )
        message = [ message, ' xK' ];
    end
    if( exit_flag_xg~=1 )
        message = [ message, ' xg' ];
    end
    disp( message );  % pause();
end


