function   results = post_processing( model_solution )
% Usage: post_processing( parameters,  model_solution )
% This function generates relevant results for a model solution.
% model_solution only needs to contain 

%% SECTION 1. Setting PARAMETERS
parameters = model_solution.parameters;
lambda=parameters.lambda;   alpha0  = parameters.alpha0 ;   beta = parameters.beta;   e = parameters.e;  theta=parameters.theta;  % Bank Run parameters.
eta = parameters.eta;  Q = parameters.Q ;   dt=parameters.dt;
PI = parameters.PI;    AL=  parameters.AL;  AH=parameters.AH;  
rho = parameters.rho;   sigmaK=parameters.sigmaK;  delta=parameters.delta;    
investment =  parameters.investment;      muK_fun = parameters.muK_fun;

%% Section 2. Generate Other Solutions
% Grid is the same for both no bankrun economy and the full model
w_grid = model_solution.w_grid;
hat_Q_vec = model_solution.hat_Q_vec;  % Important: hat Q vector. 

% Calculations for the no bankrun economy
p_ode = model_solution.p_ode45;
xg_ode =  zeros(size(w_grid));
yg_ode = (Q(w_grid)./p_ode - w_grid.*xg_ode) ./ (1-w_grid) ;   yg_ode(end) = yg_ode(end-1);

pw_ode = derivative( w_grid, p_ode );
pww_ode = derivative( w_grid, pw_ode );
psi_ode =  ( investment( p_ode ) + rho * p_ode  - AL ) ./ (AH-AL) ;
xK_ode = psi_ode ./ w_grid;  xK_ode(1) = xK_ode(2);   
yK_ode=  (1-w_grid.*xK_ode)./(1-w_grid);    yK_ode(1) = 1;   yK_ode(end)=yK_ode(end-1);
sigmap_ode = pw_ode.*w_grid.*(1-w_grid).*(xK_ode-yK_ode) ./ ( 1- pw_ode.*w_grid.*(1-w_grid).*(xK_ode-yK_ode)  ) * sigmaK;   sigmap_ode(end) = 0;
sigmab_ode = xK_ode.*(sigmaK+sigmap_ode);
sigmah_ode = yK_ode.*(sigmaK+sigmap_ode);
muR_rd_ode = (sigmaK+sigmap_ode).^2.*xK_ode  - AH./p_ode ;
rd_rg_ode =  0 ;
mub_muh_ode = ( xK_ode - yK_ode ) .* muR_rd_ode +  (xK_ode*AH-yK_ode*AL)./p_ode -  ( xg_ode - yg_ode ).*rd_rg_ode;
mup_ode = pw_ode.*w_grid.*(1-w_grid).* ( mub_muh_ode + sigmah_ode.^2 - sigmab_ode.*sigmah_ode - w_grid.*(sigmab_ode-sigmah_ode).^2 - eta./(1-w_grid) ) + 1/2*pww_ode.*( w_grid.*(1-w_grid).*(sigmab_ode-sigmah_ode) ).^2 ;   mup_ode(end) = mup_ode(end-1);
muK_ode = muK_fun(p_ode);
muR_ode = mup_ode - delta + muK_ode + sigmaK*sigmap_ode - investment(p_ode)./p_ode;
rd_ode = muR_ode - muR_rd_ode;
rg_ode = rd_ode - rd_rg_ode;
mub_ode = xK_ode.*( muR_ode + AH./p_ode - rd_ode ) - xg_ode.*rd_rg_ode + rd_ode - rho;
muh_ode = mub_ode -  mub_muh_ode;
% Evolution of w. Note: dw = muw*dt + sigmaw*dB - kappaw*dN.  They are not fractional growth parameters!
muw_ode = (1-w_grid).* ( mub_muh_ode + sigmah_ode.^2 - sigmab_ode.*sigmah_ode - w_grid.*(sigmab_ode-sigmah_ode).^2 - eta./(1-w_grid) ) ;   muw_ode(end) = muw_ode(end-1);
investment_ode = investment( p_ode );

% Calculate various quantities for the full model.
p_new = model_solution.p_new;
xg_vec = model_solution.xg_vec;
% Check if xg>=e
if( sum( xg_vec<e ) >1 )
    disp( ['Post Processing: xg>=e is not guaranteed in the model solution!'] );
end
kappap_vec = model_solution.kappap_vec;
pw_vec = derivative( w_grid, p_new );
pww_vec = derivative( w_grid, pw_vec);
psi_vec =  ( investment( p_new ) + rho * (p_new+hat_Q_vec)  - AL ) ./ (AH-AL) ;
psi_vec = min(psi_vec, 1);   % Note: sometimes hat_Q_vec is too large after one iteration. I need to supress the resulted psi_vec>1
productivity_vec = AH*psi_vec + AL*(1-psi_vec);

% Portfolio choices
xK_vec = psi_vec ./ w_grid .* p_new ./ (p_new+hat_Q_vec);  xK_vec(1) = xK_vec(2);
yK_vec = (p_new./(p_new+hat_Q_vec) -w_grid.*xK_vec)./(1-w_grid); % (1-psi_vec)./(1-w_grid) .*p_new ./ (p_new+hat_Q_vec) ;    
yK_vec(1) = 1;   yK_vec(end)=yK_vec(end-1);
yg_vec =  (Q(w_grid)./(p_new+hat_Q_vec) - w_grid.*xg_vec) ./ (1-w_grid) ;   yg_vec(end) = yg_vec(end-1);  
yg_vec = max(yg_vec, 10^(-5));  % Note: in very very rare times, the yg vec might go negative at one point. This is to eliminate the imapct. 
delta_x_vec = max(beta*(xK_vec+xg_vec-1) - xg_vec, 0);
indicator_vec = beta*(xK_vec+xg_vec-1) - xg_vec >0 ;

% Volatilities
sigmap_vec = pw_vec.*w_grid.*(1-w_grid).*(xK_vec-yK_vec) ./ ( 1- pw_vec.*w_grid.*(1-w_grid).*(xK_vec-yK_vec)  ) * sigmaK;   sigmap_vec(end) = 0;
sigmab_vec = xK_vec.*(sigmaK+sigmap_vec);
sigmah_vec = yK_vec.*(sigmaK+sigmap_vec);

% Deal with the post jump issues 
firesale_jump = (xK_vec.*kappap_vec) + alpha0/(1-alpha0)*delta_x_vec;
jump_inv = 1./(1-firesale_jump);
index = firesale_jump<0.99;
firesale_jump = interp1( w_grid(index), firesale_jump(index), w_grid, 'linear', 'extrap' );

% Jumps
kappad_vec = theta*( xK_vec+e - 1 ) ./ (xK_vec+xg_vec-1) .*(xK_vec+e>1);
kappab_vec = theta*min(1-e,xK_vec) + (1-theta)*(firesale_jump) ; 
kappafs_vec = alpha0/(1-alpha0).*delta_x_vec.*w_grid./(1-w_grid);  kappafs_vec(end)=0;
kappah_vec = yK_vec.*kappap_vec + (1-yK_vec-yg_vec).*kappad_vec - kappafs_vec ;
kappaw_vec =  (1 - (1-kappab_vec)./(1-kappah_vec-w_grid.*(kappab_vec-kappah_vec)));

% Generate liquidity premium and price of risks
liq_prem_vec = indicator_vec.*lambda*(1-PI)* (1-theta) * alpha0/(1-alpha0) ./ (1-firesale_jump) ;
index = max(find( liq_prem_vec == max( liq_prem_vec ) ) - 6, 1);
fit = slmengine( w_grid(index:end), liq_prem_vec(index:end), 'decreasing', 'on', 'rightvalue', 0,  'knots', 4 ); %, 'plot', 'on'
liq_prem_vec = [  liq_prem_vec(1:index-1) ; slmeval( w_grid(index:end), fit )];

% Liquidity holding across states.
bank_liq_frac = xg_vec.*w_grid./ (Q(w_grid)./(p_new+hat_Q_vec));

% Growth parameters. 
muR_rd_vec = (sigmaK+sigmap_vec).^2.*xK_vec + lambda*(1-theta).*( kappap_vec + indicator_vec.*alpha0/(1-alpha0)*beta ) ./ ( 1- firesale_jump ) + (xK_vec+e<1).*lambda*theta./(1-xK_vec) - AH./p_new ;

rd_rg_vec =  lambda*(1-theta)*indicator_vec*alpha0/(1-alpha0)*(1-beta) ./ ( 1 - firesale_jump ) ;
rd_rg_vec_H = lambda*kappad_vec./( 1-yK_vec.*kappap_vec-(1-yK_vec-yg_vec).*kappad_vec+kappafs_vec ) ;
index_xK_le1 = find(xK_vec<=1); % Index that xK<=1. In this scenario, the rd-rg difference must be solved from household first order condition 
rd_rg_vec( index_xK_le1 ) = rd_rg_vec_H(index_xK_le1);
mub_muh_vec = ( xK_vec - yK_vec ) .* muR_rd_vec +  (xK_vec*AH-yK_vec*AL)./p_new -  ( xg_vec - yg_vec ).*rd_rg_vec;

% Evolution of w and price.. Note: dw = muw*dt + sigmaw*dB - kappaw*dN.  They are not fractional growth parameters!
muw_vec = (1-w_grid).* ( mub_muh_vec + sigmah_vec.^2 - sigmab_vec.*sigmah_vec - w_grid.*(sigmab_vec-sigmah_vec).^2 - eta./(1-w_grid) ) ;   muw_vec(end) = muw_vec(end-1);
sigmaw_vec = (1-w_grid).*(sigmab_vec-sigmah_vec) ;
mup_vec = pw_vec.*w_grid.*muw_vec + 1/2*pww_vec.*( w_grid.*(1-w_grid).*(sigmab_vec-sigmah_vec) ).^2 ;   mup_vec(end) = mup_vec(end-1);

% Evolution of K. Note: dK/K = K_growth*dt + sigmaK*dB - kappaK*dN.  They are fractional growth parameters. 
muK_vec = muK_fun(p_new);
muR_vec = mup_vec + muK_vec - delta + sigmaK*sigmap_vec - investment(p_new)./p_new; 
K_growth_vec = muK_vec - delta ;
kappaK_vec = theta*psi_vec ;

% Other interest rates.
rd_vec = muR_vec - muR_rd_vec;
rg_vec = rd_vec - rd_rg_vec;
rd_rf_vec =  lambda*(1-theta)*indicator_vec*alpha0/(1-alpha0)*(0-beta) ./ ( 1 - firesale_jump ) ;
rf_vec = rd_vec - rd_rf_vec;  % Risk free rate for bankers.
rh_vec = rd_vec - lambda*kappad_vec./( 1-yK_vec.*kappap_vec-(1-yK_vec-yg_vec).*kappad_vec+kappafs_vec ); % Risk free rate for households.

% Other growth rates
mub_vec = xK_vec.*( muR_vec + AH./p_new - rd_vec ) - xg_vec.*rd_rg_vec + rd_vec - rho;
muh_vec = mub_vec -  mub_muh_vec;


%% Section 3. Model Output
results.w_grid = w_grid; 
results.p_new = p_new; 

results.p0 = model_solution.p0; 
results.p1 = model_solution.p1; 
results.psi_vec = psi_vec;
results.productivity_vec = productivity_vec;
results.investment_vec = investment( p_new );
results.xK_vec = xK_vec; 
results.xK_fitted = interp1(w_grid, xK_vec, 'linear', 'pp');
results.yK_vec = yK_vec; 
results.Lvg = xK_vec+xg_vec; 
results.xg_vec = xg_vec; 
results.xg_fitted = interp1(w_grid, xg_vec, 'linear', 'pp');
results.yg_vec = yg_vec; 
results.Q_vec = Q(w_grid);
results.hat_Q_vec = hat_Q_vec;
results.sigmap_vec = sigmap_vec;
results.sigmab_vec = sigmab_vec;
results.sigmah_vec = sigmah_vec;
results.kappap_vec = kappap_vec;
results.kappad_vec = kappad_vec;
results.kappah_vec = kappah_vec;
results.kappab_vec = kappab_vec;
results.kappafs_vec = kappafs_vec;
results.firesale_jump = firesale_jump;
results.kappaw_vec = kappaw_vec;
results.liq_prem_vec = liq_prem_vec;   results.liq_prem_fitted = interp1( w_grid, liq_prem_vec, 'linear', 'pp' );   
results.bank_liq_frac = bank_liq_frac;
results.delta_x_vec = delta_x_vec;
results.indicator_vec = indicator_vec;

results.mup_vec = mup_vec;
results.mub_vec = mub_vec;
results.muh_vec = muh_vec;
results.muK_vec = muK_vec;
results.muR_vec = muR_vec;
results.muR_rd_vec = muR_rd_vec;
results.mub_muh_vec = mub_muh_vec;
results.rd_vec = rd_vec;
results.rg_vec = rg_vec;
results.rf_vec = rf_vec;
results.rh_vec = rh_vec;
results.rg_fitted = interp1( w_grid, rg_vec, 'linear', 'pp');
results.rf_fitted = interp1( w_grid, rf_vec, 'linear', 'pp');

results.muw_vec = muw_vec;
results.sigmaw_vec = sigmaw_vec;

results.K_growth_vec = K_growth_vec;
results.kappaK_vec = kappaK_vec;

results.muw_fitted = interp1( w_grid,  muw_vec, 'linear', 'pp');
results.sigmaw_fitted = interp1( w_grid,  sigmaw_vec, 'linear', 'pp');
results.kappaw_fitted = interp1( w_grid,  kappaw_vec, 'linear', 'pp');

results.K_growth_fitted = interp1(   w_grid,  K_growth_vec, 'linear', 'pp' );
results.kappaK_fitted = interp1(   w_grid,  kappaK_vec, 'linear', 'pp' );

results.productivity_fitted = interp1( w_grid, productivity_vec , 'linear', 'pp' );
results.p_fitted = interp1( w_grid, p_new, 'linear', 'pp' );
results.kappap_fitted = interp1( w_grid, kappap_vec, 'linear', 'pp' );

results.p_ode = p_ode;
results.muw_ode = muw_ode;
results.muR_ode = muR_ode;
results.sigmap_ode = sigmap_ode;
results.rd_ode = rd_ode;
results.investment_ode = investment_ode;

results.errors = model_solution.errors;

% parameters
results.lambda=parameters.lambda;   results.alpha0  = parameters.alpha0 ;   results.beta = parameters.beta;   results.e = parameters.e;  results.theta=parameters.theta;  % Bank Run parameters.
results.eta = parameters.eta;  results.Q = parameters.Q ;   
results.v_prime_fun = parameters.v_prime_fun;  results.PI = parameters.PI;   
results.N_w    = parameters.N_w;   results.AL=  parameters.AL;  results.AH=parameters.AH;  results.dt=dt;
results.rho = parameters.rho;   results.sigmaK=parameters.sigmaK;  results.delta=parameters.delta;    results.chi=parameters.chi;
results.investment =  parameters.investment;    results.investmentD = parameters.investmentD;     results.muK_fun = parameters.muK_fun;



